/*
 *  This file is part of MESSIF library.
 *
 *  MESSIF library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  MESSIF library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with MESSIF library.  If not, see <http://www.gnu.org/licenses/>.
 */
package messif.objects.impl;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Map;
import messif.objects.LocalAbstractObject;
import messif.objects.MetaObject;
import messif.objects.MetaObjectParametric;
import messif.objects.keys.AbstractObjectKey;
import messif.objects.nio.BinaryInput;
import messif.objects.nio.BinaryOutput;
import messif.objects.nio.BinarySerializable;
import messif.objects.nio.BinarySerializator;

/**
 * Implementation of the {@link MetaObject} that stores a fixed array of encapsulated objects
 * with additional {@link messif.utility.Parametric parameters}.
 * The metric distance function for this object is the absolute value of the
 * differences of locatorURI hash codes. For a more sophisticated distance function
 * use {@link MetaObjectArrayWeightedSum}.
 * 
 * <p>
 * Note that the encapsulated object names are automatically generated by
 * {@link #getObjectName(int)} method. Normally, this method is overloaded
 * in a subclass to provide more appropriate names.
 * </p>
 * 
 * @author Michal Batko, Masaryk University, Brno, Czech Republic, batko@fi.muni.cz
 * @author Vlastislav Dohnal, Masaryk University, Brno, Czech Republic, dohnal@fi.muni.cz
 * @author David Novak, Masaryk University, Brno, Czech Republic, david.novak@fi.muni.cz
 */
public class MetaObjectParametricArray extends MetaObjectParametric implements BinarySerializable {
    /** Class id for serialization. */
    private static final long serialVersionUID = 1L;

    //****************** Attributes ******************//

    /** List of objects */
    protected final LocalAbstractObject[] objects;


    //****************** Constructors ******************//

    /**
     * Creates a new instance of MetaObjectArray.
     * A new unique object ID is generated and the
     * object's key is set to <tt>null</tt>.
     * @param additionalParameters additional parameters for this meta object
     * @param objects the encapsulated list of objects
     */
    public MetaObjectParametricArray(Map<String, ? extends Serializable> additionalParameters, LocalAbstractObject... objects) {
        super(additionalParameters);
        this.objects = objects.clone();
    }

    /**
     * Creates a new instance of MetaObjectArray.
     * A new unique object ID is generated and the 
     * object's key is set to the specified key.
     * @param objectKey the key to be associated with this object
     * @param additionalParameters additional parameters for this meta object
     * @param objects the encapsulated list of objects
     */
    public MetaObjectParametricArray(AbstractObjectKey objectKey, Map<String, ? extends Serializable> additionalParameters, LocalAbstractObject... objects) {
        super(objectKey, additionalParameters);
        this.objects = objects.clone();
    }

    /**
     * Creates a new instance of MetaObjectArray.
     * A new unique object ID is generated and a
     * new {@link AbstractObjectKey} is generated for
     * the specified <code>locatorURI</code>.
     * @param locatorURI the locator URI for the new object
     * @param additionalParameters additional parameters for this meta object
     * @param objects the encapsulated list of objects
     */
    public MetaObjectParametricArray(String locatorURI, Map<String, ? extends Serializable> additionalParameters, LocalAbstractObject... objects) {
        super(locatorURI, additionalParameters);
        this.objects = objects.clone();
    }

    /**
     * Creates a new instance of MetaObjectArray that takes the objects from the given collection.
     * A new unique object ID is generated and a new {@link AbstractObjectKey} is
     * generated for the specified <code>locatorURI</code>.
     * @param locatorURI the locator URI for the new object
     * @param additionalParameters additional parameters for this meta object
     * @param objects the collection with objects to encapsulate
     */
    public MetaObjectParametricArray(String locatorURI, Map<String, ? extends Serializable> additionalParameters, Collection<? extends LocalAbstractObject> objects) {
        super(locatorURI, additionalParameters);
        this.objects = objects.toArray(new LocalAbstractObject[objects.size()]);
    }

    /**
     * Creates a new instance of MetaObjectArray that takes the objects from the given map.
     * The array is initialized with objects from the map in the order they
     * appear in the {@code objectNames} array. Note that if the object of a given
     * name is not in the map, <tt>null</tt> is inserted into the array.
     * A new unique object ID is generated and a new {@link AbstractObjectKey} is
     * generated for the specified <code>locatorURI</code>.
     * @param locatorURI the locator URI for the new object
     * @param additionalParameters additional parameters for this meta object
     * @param objects the map with named objects to encapsulate
     * @param objectNames the names of the objects to take from the given {@code objects} map
     */
    public MetaObjectParametricArray(String locatorURI, Map<String, ? extends Serializable> additionalParameters, Map<String, ? extends LocalAbstractObject> objects, String... objectNames) {
        this(new AbstractObjectKey(locatorURI), additionalParameters, objects, objectNames);
    }

    /**
     * Creates a new instance of MetaObjectArray that takes the objects from the given map.
     * The array is initialized with objects from the map in the order they
     * appear in the {@code objectNames} array. Note that if the object of a given
     * name is not in the map, <tt>null</tt> is inserted into the array.
     * A new unique object ID is generated and the given {@code objectKey} is used for the object.
     * @param objectKey the key to be associated with this object
     * @param additionalParameters additional parameters for this meta object
     * @param objects the map with named objects to encapsulate
     * @param objectNames the names of the objects to take from the given {@code objects} map
     */
    public MetaObjectParametricArray(AbstractObjectKey objectKey, Map<String, ? extends Serializable> additionalParameters, Map<String, ? extends LocalAbstractObject> objects, String... objectNames) {
        super(objectKey, additionalParameters);
        this.objects = new LocalAbstractObject[objectNames.length];
        for (int i = 0; i < objectNames.length; i++)
            this.objects[i] = objects.get(objectNames[i]);
    }

    /**
     * Creates a new instance of MetaObjectArray from the given text stream.
     * @param stream the text stream to read the objects from
     * @param additionalParameters additional parameters for this meta object
     * @param classes the classes of the objects to read from the stream
     * @throws IOException when an error appears during reading from given stream,
     *         EOFException is returned if end of the given stream is reached.
     */
    public MetaObjectParametricArray(BufferedReader stream, Map<String, ? extends Serializable> additionalParameters, Class<? extends LocalAbstractObject>[] classes) throws IOException {
        this(stream, additionalParameters, false, classes);
    }

    /**
     * Creates a new instance of MetaObjectArray from the given text stream.
     * @param stream the text stream to read the objects from
     * @param additionalParameters additional parameters for this meta object
     * @param readEmptyClassLine flag whether to read a single line (that is ignored)
     *          for a <tt>null</tt> class inside the {@code classes array}
     * @param classes the classes of the objects to read from the stream
     * @throws IOException when an error appears during reading from given stream,
     *         EOFException is returned if end of the given stream is reached.
     */
    public MetaObjectParametricArray(BufferedReader stream, Map<String, ? extends Serializable> additionalParameters, boolean readEmptyClassLine, Class<? extends LocalAbstractObject>[] classes) throws IOException {
        super(additionalParameters);
        readObjectCommentsWithoutData(stream);
        this.objects = MetaObjectArray.readObjects(stream, readEmptyClassLine, classes);
    }

    /**
     * Creates a new instance of MetaObjectArray from the given text stream.
     * @param stream the text stream to read the objects from
     * @param additionalParameters additional parameters for this meta object
     * @param objectCount number of objects to read
     * @param objectClass the class of objects to read from the stream
     * @throws IOException when an error appears during reading from given stream,
     *         EOFException is returned if end of the given stream is reached.
     */
    public MetaObjectParametricArray(BufferedReader stream, Map<String, ? extends Serializable> additionalParameters, int objectCount, Class<? extends LocalAbstractObject> objectClass) throws IOException {
        this(stream, additionalParameters, MetaObjectArray.createClassArray(objectCount, objectClass));
    }

    /**
     * Creates a new instance of MetaObjectArray from the given text stream.
     * @param stream the text stream to read the objects from
     * @param classes the classes of the objects to read from the stream
     * @throws IOException when an error appears during reading from given stream,
     *         EOFException is returned if end of the given stream is reached.
     */
    public MetaObjectParametricArray(BufferedReader stream, Class<? extends LocalAbstractObject>[] classes) throws IOException {
        this(stream, null, classes);
    }

    /**
     * Creates a new instance of MetaObjectArray from the given text stream with header.
     * Note that a header must contain also the object names even though they are not
     * stored and used by the array.
     * @param stream the text stream to read the objects from
     * @throws IOException when an error appears during reading from given stream,
     *         EOFException is returned if end of the given stream is reached.
     * @see #readObjectsHeader(java.io.BufferedReader)
     */
    public MetaObjectParametricArray(BufferedReader stream) throws IOException {
        super(null);
        this.objects = readObjects(stream, null, readObjectsHeader(stream), new LinkedHashMap<String, LocalAbstractObject>()).values().toArray(new LocalAbstractObject[0]);
    }


    //****************** MetaObject method implementations ******************//

    /**
     * Returns the name of the fixed object with the given {@code index}.
     * By default, this implementation returns "ObjectX" where X is the given
     * {@code index}. However, this method should be overridden to give
     * the real name of the object if appropriate (e.g. from a static array of names).
     * 
     * @param index the fixed index of the object the name of which to get
     * @return the name of the {@code index}th object
     */
    protected String getObjectName(int index) {
        return "Object" + index;
    }


    @Override
    public int getObjectCount() {
        int count = 0;
        for (LocalAbstractObject object : objects) {
            if (object != null) {
                count++;
            }
        }
        return count;
    }

    @Override
    public LocalAbstractObject getObject(String name) {
        for (int i = 0; i < objects.length; i++)
            if (getObjectName(i).equals(name))
                return objects[i];
        return null;
    }

    /**
     * Returns the encapsulated object for given index.
     *
     * @param index the index of the object to return
     * @return encapsulated object for given name or <tt>null</tt> if the key is unknown
     * @throws IndexOutOfBoundsException if the given index is not within [0;{@link #getObjectCount() count}) interval
     */
    public LocalAbstractObject getObject(int index) {
        return objects[index];
    }

    @Override
    public Collection<String> getObjectNames() {
        Collection<String> names = new ArrayList<String>(objects.length);
        for (int i = 0; i < objects.length; i++)
            if (objects[i] != null)
                names.add(getObjectName(i));
        return names;
    }

    @Override
    public Collection<LocalAbstractObject> getObjects() {
        Collection<LocalAbstractObject> ret = new ArrayList<LocalAbstractObject>(objects.length);
        for (LocalAbstractObject object : objects) {
            if (object != null) {
                ret.add(object);
            }
        }
        return ret;
    }

    @Override
    public Map<String, LocalAbstractObject> getObjectMap() {
        Map<String, LocalAbstractObject> ret = new LinkedHashMap<String, LocalAbstractObject>(objects.length);
        for (int i = 0; i < objects.length; i++) {
            if (objects[i] != null)
                ret.put(getObjectName(i), objects[i]);
        }
        return ret;
    }


    //****************** Text stream I/O helper method ******************//

    @Override
    protected void writeDataImpl(OutputStream stream) throws IOException {
        for (LocalAbstractObject object : objects) {
            if (object == null) {
                stream.write('\n');
            } else {
                object.write(stream, false);
            }
        }
    }


    //****************** Distance function ******************//

    /**
     * The actual implementation of the metric function.
     * The distance is a trivial metric on locator URIs of this object and {@code obj}.
     * The array <code>metaDistances</code> is ignored.
     * @return the actual distance between obj and this if the distance is lower than distThreshold
     * @see LocalAbstractObject#getDistance(messif.objects.LocalAbstractObject, float) LocalAbstractObject.getDistance
     */
    @Override
    protected float getDistanceImpl(LocalAbstractObject obj, float[] metaDistances, float distThreshold) {
        String thisLocator = getLocatorURI();
        String otherLocator = obj.getLocatorURI();
        return (thisLocator == null ? otherLocator == null : thisLocator.equals(otherLocator)) ? 0 : 1;
    }


    //************ BinarySerializable support ************//

    /**
     * Creates a new instance of MetaObjectFixed loaded from binary input.
     * 
     * @param input the input to read the MetaObject from
     * @param serializator the serializator used to write objects
     * @throws IOException if there was an I/O error reading from the buffer
     */
    protected MetaObjectParametricArray(BinaryInput input, BinarySerializator serializator) throws IOException {
        super(input, serializator);
        this.objects = new LocalAbstractObject[serializator.readInt(input)];
        for (int i = 0; i < this.objects.length; i++)
            this.objects[i] = serializator.readObject(input, LocalAbstractObject.class);
    }

    @Override
    public int binarySerialize(BinaryOutput output, BinarySerializator serializator) throws IOException {
        int size = super.binarySerialize(output, serializator);
        size += serializator.write(output, objects.length);
        for (LocalAbstractObject object : objects) {
            size += serializator.write(output, object);
        }
        return size;
    }

    @Override
    public int getBinarySize(BinarySerializator serializator) {
        int size = super.getBinarySize(serializator);
        size += serializator.getBinarySize(objects.length);
        for (LocalAbstractObject object : objects) {
            size += serializator.getBinarySize(object);
        }
        return size;
    }

}
